#include "robots.h"
#include "level.h"
#include "palette.h"
#include "game.h"
#include "toppler.h"

#include <stdlib.h>

/* this field contains the robots */
static struct {

  /* the position of the robot */
  int anglepos;
  long verticalpos;

  /* what kind of robot it is, an under classifikation
   and what kind it will be after the appearing animation */
  int kind;
  long subKind;
  int futureKind;

  /* a timer for the animations of the robots */
  long time;
} objekt[4];


/* the position up to where the the robots are worked out */
static int robots_ready;
static int robots_angle;

/* the data for the next cross that will appear */
static int next_cross_timer;
static int cross_direction;
static int nextcrosscolor;


/******** PRIVATE FUNCTIONS ********/

/* returns the index of the figure the given figure (nr) kollides
 with or -1 if there is no such object */
static int figurkollision(int nr) {

  /* help field for kollisions between two objects */
  static unsigned char kollision[16] = {
    0x1c, 0x3e, 0x3e, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x3e, 0x3e, 0x1c, 0x00
  };

  int t;
  long i, j;

  for (t = 0; t <= 3; t++) {
    if (t != nr &&
        objekt[t].kind != OBJ_KIND_CROSS &&
        objekt[t].kind != OBJ_KIND_NOTHING &&
        objekt[t].kind != OBJ_KIND_DISAPPEAR &&
        objekt[t].kind != OBJ_KIND_APPEAR) {
      i = objekt[nr].anglepos - objekt[t].anglepos;
      j = objekt[t].verticalpos - objekt[nr].verticalpos;
      if ((-4 < i) && (i < 4)  && (-8 < j) && (j < 8)) {
        if ((kollision[j + 7] >> (i + 3)) & 1)
          return t;
        else
          return -1;
      }
    }
  }
  return -1;
}

/* returns true, if the roboter can be at the given position without colliding
 with an element from the tower */
static bool testroboter(int nr) {
  return (!lev_testfigure((long)objekt[nr].anglepos, objekt[nr].verticalpos, -2L, 1L, 1L, 1L, 7L));
}

/* makes the robot disappear when it falle into water */
static void drown_robot(int nr) {
  objekt[nr].verticalpos = 0;
  objekt[nr].time = 0;
  objekt[nr].kind = OBJ_KIND_DISAPPEAR;
}

/* tests the undergroud of the given object (only used for freeze ball) returns
 0 if nothing needs to be done
 1 for falling
 2 for reverse direction
 */
static int test_robot_undergr(int nr) {

  int row, col;
  row = objekt[nr].verticalpos / 4 - 1;
  col = (objekt[nr].anglepos / 8) & 0xf;

  if (lev_is_box(row, col)) return 0;
  if (lev_is_platform(row, col) || lev_is_stick(row, col)) {
    if (lev_is_elevator(row, col))
      return 2;
    else
      return 0;
  }

  if ((objekt[nr].anglepos & 7) < 2) {
    if (lev_is_empty(row, (col - 1) & 0xf)) return 1;
    if (lev_is_door(row, (col - 1) & 0xf)) return 1;
    if ((objekt[nr].subKind & 0x80) == 0)
      return 2;
    else
      return 0;
  }

  if ((objekt[nr].anglepos & 7) > 6) {
    if (lev_is_empty(row, (col + 1) & 0xf)) return 1;
    if (lev_is_door(row, (col + 1) & 0xf)) return 1;
    if ((objekt[nr].subKind & 0x80) == 0)
      return 0;
    else
      return 2;
  }

  return 1;
}

/* update the position for the cross */
static void updatecross(int t) {

  objekt[t].anglepos += objekt[t].subKind;
  objekt[t].time -= objekt[t].subKind;

  /* after the cross reached the middle speed it up */
  if (objekt[t].anglepos == 60)
    objekt[t].subKind *= 2;

  /* if cross reached screenedge, remove it an reinitialize
   counter for next one */
  if (((objekt[t].subKind > 0) && (objekt[t].anglepos >= 120)) ||
      ((objekt[t].subKind < 0) && (objekt[t].anglepos <= 0))) {
    objekt[t].kind = OBJ_KIND_NOTHING;
    next_cross_timer = 125;
  }
}

/* remove objects that drop below the screen */
static bool checkverticalposition(int verticalpos, int t) {

  if (objekt[t].verticalpos + 48 < verticalpos) {
    objekt[t].kind = OBJ_KIND_DISAPPEAR;
    objekt[t].time = 0;
    return true;
  } else
    return false;

}

/* checks if the robot is at a position that it can not be
 if not remove it */
static bool checkvalidposition(int t) {

  if (testroboter(t)) {
    objekt[t].kind = OBJ_KIND_DISAPPEAR;
    objekt[t].time = 0;
    return true;
  } else
    return false;

}

/* move the robot a the horizontal amount, if this places it
 at an impossible position, reverse its direction and don't move it */
static void moverobothorizontal(int t) {

  objekt[t].anglepos += objekt[t].subKind;
  objekt[t].anglepos &= 0x7f;
  if (testroboter(t) || (figurkollision(t) != -1)) {
    objekt[t].subKind = -objekt[t].subKind;
    objekt[t].anglepos += objekt[t].subKind;
    objekt[t].anglepos &= 0x7f;
  }

}


/******* PUBLIC FUNCTIONS *********/

int rob_kind(int nr) { return objekt[nr].kind; }
int rob_time(int nr) { return objekt[nr].time; }
int rob_angle(int nr) { return objekt[nr].anglepos; }
int rob_vertical(int nr) { return objekt[nr].verticalpos; }

void rob_initialize() {

  for (int b = 0; b <= 4; b++) {
    objekt[b].kind = OBJ_KIND_NOTHING;
    objekt[b].time = -1;
  }

  next_cross_timer = 125;
  nextcrosscolor = rand() / (RAND_MAX / 8);
  cross_direction = 1;

  robots_ready = 0;
  robots_angle = 0;

}

int rob_topplerkollision(int angle, int vertical) {
  long i, j;

  for (int t = 0; t <= 3; t++) {
    if (objekt[t].kind != OBJ_KIND_CROSS &&
        objekt[t].kind != OBJ_KIND_NOTHING &&
        objekt[t].kind != OBJ_KIND_DISAPPEAR &&
        objekt[t].kind != OBJ_KIND_APPEAR) {
      i = angle - objekt[t].anglepos;
      j = objekt[t].verticalpos - vertical;
      if ((-4 < i) && (i < 4)  && (-8 < j) && (j < 8))
        return t;
    } else if (objekt[t].kind == OBJ_KIND_CROSS) {
      if (objekt[t].anglepos >= 53 && objekt[t].anglepos <= 67 &&
          objekt[t].verticalpos + 7 >= vertical &&
          vertical + 9 >= objekt[t].verticalpos)
        return t;
    }
  }
  return -1;
}

int rob_snowballkollision(int angle, int vertical) {

  /* help field for kollisions between two objects */
  static unsigned char kollision[16] = {
     0x00, 0x1c, 0x1c, 0x3e, 0x3e, 0x3e, 0x3e, 0x3e, 0x3e, 0x3e, 0x1c, 0x1c, 0x00, 0x00, 0x00, 0x10
  };

  long i, j;
  for (int t = 0; t <= 3; t++) {
    if (objekt[t].kind != OBJ_KIND_CROSS &&
        objekt[t].kind != OBJ_KIND_NOTHING &&
        objekt[t].kind != OBJ_KIND_DISAPPEAR &&
        objekt[t].kind != OBJ_KIND_APPEAR) {
      i = angle - objekt[t].anglepos;
      j = objekt[t].verticalpos - vertical;
      if ((-4 < i) && (i < 4)  && (-8 < j) && (j < 8)) {
        if ((kollision[j + 7] >> (i + 3)) & 1)
          return t;
        else
          return -1;
      }
    }
  }
  return -1;
}

void rob_new(int verticalpos) {

  static struct {
    unsigned char r, g, b;
  } crosscols[8] = {
    { 0xff, 0x00, 0x00 },
    { 0x00, 0xff, 0x00 },
    { 0x00, 0x00, 0xff },
    { 0xff, 0xff, 0x00 },
    { 0x00, 0xff, 0xff },
    { 0xff, 0x00, 0xff },
    { 0x80, 0x80, 0x80 },
    { 0xff, 0xff, 0xff }
  };

  int a, b;

  int h = verticalpos / 4 + 9;

  for (int t = 0; t <= 3; t++) {
    if (objekt[t].kind == OBJ_KIND_NOTHING) {


      /* if there is currently no chance for a new robot check for a cross */
      if ((robots_ready == lev_towerheight()) || (h <= robots_ready)) {

        /* check if time is ripe */
        if (next_cross_timer == -1) return;
        next_cross_timer--;
        if (next_cross_timer != -1) return;

        /* set colors for the cross */
        pal_setcrosscolors(crosscols[nextcrosscolor].r,
                           crosscols[nextcrosscolor].g,
                           crosscols[nextcrosscolor].b);
        nextcrosscolor = (nextcrosscolor + 1) & 7;

        /* fill in the data for the robot */
        objekt[t].kind = OBJ_KIND_CROSS;
        objekt[t].time = 0;
        cross_direction *= -1;
        objekt[t].subKind = cross_direction;
        if (cross_direction > 0)
          objekt[t].anglepos = 0;
        else
          objekt[t].anglepos = 120;
        objekt[t].verticalpos = verticalpos;

      } else {

        /* find next robot */
        do {
          b = lev_tower(robots_ready, robots_angle);
          a = robots_angle;
          h = robots_ready;
          robots_angle = (robots_angle + 1) & 0xf;
        } while (((b & 0x80) != 0 || (b & 0x70) == 0) && robots_angle != 0);
  
        if (robots_angle == 0)
          robots_ready++;
  
        /* no robot found */
        if ((b & 0x80) != 0 || (b & 0x70) == 0) return;
  
        /* fill in data for robot */
        switch (b) {
  
          case 0x10:
            objekt[t].subKind = 1;
            objekt[t].futureKind = OBJ_KIND_FREEZEBALL;
            break;
    
          case 0x20:
            objekt[t].subKind = 1;
            objekt[t].futureKind = OBJ_KIND_JUMPBALL;
            break;
    
          case 0x30:
            objekt[t].subKind = 0;
            objekt[t].futureKind = OBJ_KIND_JUMPBALL;
            break;
    
          case 0x40:
            objekt[t].subKind = 1;
            objekt[t].futureKind = OBJ_KIND_ROBOT_VERT;
            break;
    
          case 0x50:
            objekt[t].subKind = 2;
            objekt[t].futureKind = OBJ_KIND_ROBOT_VERT;
            break;
    
          case 0x60:
            objekt[t].subKind = 1;
            objekt[t].futureKind = OBJ_KIND_ROBOT_HORIZ;
            break;
    
          case 0x70:
            objekt[t].subKind = 2;
            objekt[t].futureKind = OBJ_KIND_ROBOT_HORIZ;
            break;
        }
        objekt[t].anglepos = (a * 8) + 4;
        objekt[t].verticalpos = h * 4;
        objekt[t].kind = OBJ_KIND_APPEAR;
        objekt[t].time = 0;
  
        /* empty the field in the level datastructure */
        lev_clear(h, a);
      }

      return;
    }
  }
}

void rob_aktualize() {

  /* the vertical movement of the jumping ball */
  static long springball[11] = {
    2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -4
  };

  int h;

  for (int t = 0; t <= 3; t++) {
    switch (objekt[t].kind) {
      case OBJ_KIND_NOTHING:
        break;

      case OBJ_KIND_CROSS:
        updatecross(t);
        break;

      case OBJ_KIND_DISAPPEAR:
        if (objekt[t].time == 6)
          objekt[t].kind = OBJ_KIND_NOTHING;
        else
          objekt[t].time++;
        break;

      case OBJ_KIND_APPEAR:
        if (objekt[t].time == 6) {
          objekt[t].kind = objekt[t].futureKind;
          objekt[t].time = 0;
        } else
          objekt[t].time++;
        break;

      case OBJ_KIND_FREEZEBALL_FROZEN:
        objekt[t].time--;
        if (objekt[t].time > 0)
          break;
        objekt[t].kind = OBJ_KIND_FREEZEBALL;

      case OBJ_KIND_FREEZEBALL:

        if (checkverticalposition(top_verticalpos(), t) ||
            checkvalidposition(t)) break;

        switch (test_robot_undergr(t)) {
          case 1:
            objekt[t].kind = OBJ_KIND_FREEZEBALL_FALLING;
            objekt[t].time = 10;
            break;
          case 2:
            objekt[t].subKind = -objekt[t].subKind;
            break;
        }
        moverobothorizontal(t);
        break;

      case OBJ_KIND_FREEZEBALL_FALLING:

        if (checkverticalposition(top_verticalpos(), t) ||
            checkvalidposition(t)) break;

        moverobothorizontal(t);

        if (objekt[t].verticalpos + springball[objekt[t].time] < 0) {
          drown_robot(t);
          break;
        }

        h = springball[objekt[t].time];

        while (h != 0) {

          objekt[t].verticalpos += h;
          if (testroboter(t) | (figurkollision(t) != -1)) break;
          objekt[t].verticalpos -= h;

          if (h > 0)
            h--;
          else
            h++;
        }

        if ((h == 0) && (springball[objekt[t].time] < 0)) {
          objekt[t].kind = OBJ_KIND_FREEZEBALL;
          objekt[t].time = 0;
        }
        break;

      case OBJ_KIND_JUMPBALL:

        if (checkverticalposition(top_verticalpos(), t) ||
            checkvalidposition(t)) break;

        moverobothorizontal(t);

        if (objekt[t].verticalpos + springball[objekt[t].time] < 0) {
          drown_robot(t);
          break;
        }

        h = springball[objekt[t].time];

        while (h != 0) {

          objekt[t].verticalpos += h;
          if (!testroboter(t) || (figurkollision(t) != -1)) break;
          objekt[t].verticalpos -= h;

          if (h > 0)
            h--;
          else
            h++;
        }

        /* ball is bouncing */
        if ((h == 0) && (springball[objekt[t].time] < 0)) {

          /* restart bounce cyclus */
          objekt[t].time = 0;

          /* start the bounding ball moving sideways in direction of the animal */
          if (objekt[t].subKind == 0 && top_visible() && top_walking() &&
              objekt[t].verticalpos == top_verticalpos()) {

            /* check if the animal is near enough to the ball */
            unsigned char w = (objekt[t].anglepos - top_anglepos()) & 0x7f;
            h = -1;
            if (w >= 0x40)
              w |= 0x80;
            if (w >= 0x80) {
              w = 0xff & (~w + 1);
              h = 1;
            }

            if (w < 0x20) objekt[t].subKind = h;
          }
          break;
        }

        objekt[t].time++;
        if (objekt[t].time > 10)
          objekt[t].time = 10;

        break;

      case OBJ_KIND_ROBOT_HORIZ:

        if (checkverticalposition(top_verticalpos(), t) ||
            checkvalidposition(t)) break;

        moverobothorizontal(t);

        objekt[t].time = (objekt[t].time + 1) & 0x1f;
        break;

      case OBJ_KIND_ROBOT_VERT:

        if (checkverticalposition(top_verticalpos(), t) ||
            checkvalidposition(t)) break;

        if ((objekt[t].verticalpos + objekt[t].subKind) < 0)
          drown_robot(t);
        else {
          objekt[t].verticalpos += objekt[t].subKind;

          if (testroboter(t) || (figurkollision(t) != -1)) {
            objekt[t].subKind *= -1;
            objekt[t].verticalpos += objekt[t].subKind;
          }
        }

        objekt[t].time = (objekt[t].time + 1) & 0x1f;
        break;
    }
  }
}

int rob_gothit(int nr) {

  if (objekt[nr].kind == OBJ_KIND_FREEZEBALL) {
    objekt[nr].time = 0x4b;
    objekt[nr].kind = OBJ_KIND_FREEZEBALL_FROZEN;
    return 0;
  } else if (objekt[nr].kind == OBJ_KIND_JUMPBALL) {
    objekt[nr].kind = OBJ_KIND_DISAPPEAR;
    objekt[nr].time = 0;
    return 100;
  } else
    return 0;

}

void rob_disappearall() {

  for (int t = 0; t <= 3; t++) {
    if (objekt[t].kind != OBJ_KIND_NOTHING) {
      objekt[t].kind = OBJ_KIND_DISAPPEAR;
      objekt[t].time = 0;
    }
  }

}

